# encoding: utf-8
# This file is distributed under Mmtrix's license terms.


require File.expand_path(File.join(File.dirname(__FILE__),'..','..','..','test_helper'))
require 'mmtrix/agent/configuration/environment_source'

module Mmtrix::Agent::Configuration
  class EnvironmentSourceTest < Minitest::Test

    def setup
      @original_env = {}
      @original_env.replace(ENV)
      @environment_source = EnvironmentSource.new
    end

    def teardown
      ENV.replace(@original_env)
    end

    def test_environment_strings_are_applied
      assert_applied_string 'NRCONFIG', 'config_path'
      assert_applied_string 'MM_TRIX_LICENSE_KEY', 'license_key'
      assert_applied_string 'MMTRIX_LICENSE_KEY', 'license_key'
      assert_applied_string 'MM_TRIX_APP_NAME', 'app_name'
      assert_applied_string 'MMTRIX_APP_NAME', 'app_name'
      assert_applied_string 'MM_TRIX_HOST', 'host'
    end

    def test_environment_fixnums_are_applied
      assert_applied_fixnum 'MM_TRIX_PORT', 'port'
    end

    def test_environment_symbols_are_applied
      assert_applied_symbol 'MM_TRIX_DISPATCHER', 'dispatcher'
      assert_applied_symbol 'MMTRIX_DISPATCHER', 'dispatcher'
      assert_applied_symbol 'MM_TRIX_FRAMEWORK', 'framework'
      assert_applied_symbol 'MMTRIX_FRAMEWORK', 'framework'
    end

    %w| MMTRIX_ENABLE MMTRIX_ENABLED MM_TRIX_ENABLE MM_TRIX_ENABLED |.each do |var|
      define_method("test_environment_booleans_truths_are_applied_to_#{var}") do
        ENV[var] = 'true'
        assert EnvironmentSource.new[:enabled]
        ENV[var] = 'on'
        assert EnvironmentSource.new[:enabled]
        ENV[var] = 'yes'
        assert EnvironmentSource.new[:enabled]
        ENV.delete(var)
      end

      define_method("test_environment_booleans_falsehoods_are_applied_to_#{var}") do
        ENV[var] = 'false'
        assert !EnvironmentSource.new[:enabled]
        ENV[var] = 'off'
        assert !EnvironmentSource.new[:enabled]
        ENV[var] = 'no'
        assert !EnvironmentSource.new[:enabled]
        ENV.delete(var)
      end
    end

    %w| MMTRIX_DISABLE_HARVEST_THREAD MM_TRIX_DISABLE_HARVEST_THREAD |.each do |var|
      define_method("test_environment_booleans_truths_are_applied_to_#{var}") do
        ENV[var] = 'true'
        assert EnvironmentSource.new[:disable_harvest_thread]
        ENV[var] = 'on'
        assert EnvironmentSource.new[:disable_harvest_thread]
        ENV[var] = 'yes'
        assert EnvironmentSource.new[:disable_harvest_thread]
        ENV.delete(var)
      end

      define_method("test_environment_booleans_falsehoods_are_applied_to_#{var}") do
        ENV[var] = 'false'
        assert !EnvironmentSource.new[:disable_harvest_thread]
        ENV[var] = 'off'
        assert !EnvironmentSource.new[:disable_harvest_thread]
        ENV[var] = 'no'
        assert !EnvironmentSource.new[:disable_harvest_thread]
        ENV.delete(var)
      end
    end

    def test_set_log_config_from_environment
      ENV['MM_TRIX_LOG'] = 'off/in/space.log'
      source = EnvironmentSource.new
      assert_equal 'off/in', source[:log_file_path]
      assert_equal 'space.log', source[:log_file_name]
    end

    def test_set_log_config_STDOUT_from_environment
      ENV['MM_TRIX_LOG'] = 'STDOUT'
      source = EnvironmentSource.new
      assert_equal 'STDOUT', source[:log_file_name]
      assert_equal 'STDOUT', source[:log_file_path]
    end

    def test_set_values_from_mmtrix_environment_variables
      keys = %w(MM_TRIX_LICENSE_KEY MMTRIX_CONFIG_PATH)
      keys.each { |key| ENV[key] = 'skywizards' }

      expected_source = EnvironmentSource.new

      [:license_key, :config_path].each do |key|
        assert_equal 'skywizards', expected_source[key]
      end
    end

    def test_set_values_from_mmtrix_environment_variables_warns_unknowns
      ENV['MMTRIX_DOESNT_USE_THIS_VALUE'] = "true"
      expects_logging(:info, includes("MMTRIX_DOESNT_USE_THIS_VALUE"))
      @environment_source.set_values_from_mmtrix_environment_variables
    end

    def test_set_values_from_mmtrix_environment_variables_ignores_MM_TRIX_LOG
      ENV['MM_TRIX_LOG'] = "STDOUT"
      expects_no_logging(:info)
      @environment_source.set_values_from_mmtrix_environment_variables
    end

    def test_set_value_from_environment_variable
      ENV['MM_TRIX_LICENSE_KEY'] = 'super rad'
      @environment_source.set_value_from_environment_variable('MM_TRIX_LICENSE_KEY')
      assert_equal @environment_source[:license_key], 'super rad'
    end

    def test_set_key_by_type_uses_the_default_type
      ENV['MM_TRIX_TEST'] = 'true'
      @environment_source.set_key_by_type(:enabled, 'MM_TRIX_TEST')
      assert_equal true, @environment_source[:enabled]
    end

    def test_set_key_with_mmtrix_prefix
      assert_applied_string('MM_TRIX_LICENSE_KEY', :license_key)
    end

    def test_set_key_with_mmtrix_prefix
      assert_applied_string('MMTRIX_LICENSE_KEY', :license_key)
    end

    def test_does_not_set_key_without_mmtrix_related_prefix
      ENV['CONFIG_PATH'] = 'boom'
      refute_equal 'boom', EnvironmentSource.new[:config_path]
    end

    def test_convert_environment_key_to_config_key
      result = @environment_source.convert_environment_key_to_config_key('MM_TRIX_IS_RAD')
      assert_equal :is_rad, result
    end

    def test_convert_environment_key_to_config_key_respects_aliases
      assert_applied_boolean('MMTRIX_ENABLE', :enabled)
    end

    def test_convert_environment_key_to_config_key_allows_underscores_as_dots
      assert_applied_string('MM_TRIX_AUDIT_LOG_PATH', :'audit_log.path')
    end

    def test_collect_mmtrix_environment_variable_keys
      keys = %w(MM_TRIX_IS_RAD MMTRIX_IS_MAGIC)
      keys.each { |key| ENV[key] = 'true' }

      result = @environment_source.collect_mmtrix_environment_variable_keys
      assert_equal keys, result
    end

    def test_does_not_warn_for_mmtrix_env_environment_variable
      expects_no_logging(:warn)
      expects_no_logging(:info)
      with_environment('MM_TRIX_ENV' => 'foo') do
        @environment_source.set_values_from_mmtrix_environment_variables
      end
    end

    def assert_applied_string(env_var, config_var)
      ENV[env_var] = 'test value'
      assert_equal 'test value', EnvironmentSource.new[config_var.to_sym]
      ENV.delete(env_var)
    end

    def assert_applied_symbol(env_var, config_var)
      ENV[env_var] = 'test value'
      assert_equal :'test value', EnvironmentSource.new[config_var.to_sym]
      ENV.delete(env_var)
    end

    def assert_applied_fixnum(env_var, config_var)
      ENV[env_var] = '3000'
      assert_equal 3000, EnvironmentSource.new[config_var.to_sym]
      ENV.delete(env_var)
    end

    def assert_applied_boolean(env_var, config_var)
      ENV[env_var] = 'true'
      assert_equal true, EnvironmentSource.new[config_var.to_sym]
      ENV.delete(env_var)
    end

  end
end
