# encoding: utf-8
# This file is distributed under Mmtrix's license terms.


require File.expand_path(File.join(File.dirname(__FILE__),'..', '..','..','test_helper'))
require 'mmtrix/agent/samplers/cpu_sampler'

class Mmtrix::Agent::StatsEngine::SamplersTest < Minitest::Test

  class OurSamplers
    include Mmtrix::Agent::StatsEngine::Samplers
  end

  class OurSampler
    attr_accessor :id, :stats_engine
  end

  def setup
    @stats_engine = Mmtrix::Agent::StatsEngine.new
    Mmtrix::Agent.instance.stubs(:stats_engine).returns(@stats_engine)
  end

  def test_cpu_sampler_records_user_and_system_time
    timeinfo0 = mock
    timeinfo0.stubs(:utime).returns(10.0)
    timeinfo0.stubs(:stime).returns(5.0)

    timeinfo1 = mock
    timeinfo1.stubs(:utime).returns(14.0) # +5s
    timeinfo1.stubs(:stime).returns(7.0)  # +2s

    elapsed = 10

    freeze_time
    Process.stubs(:times).returns(timeinfo0, timeinfo1)
    Mmtrix::Agent::SystemInfo.stubs(:num_logical_processors).returns(4)

    s = Mmtrix::Agent::Samplers::CpuSampler.new # this calls poll
    advance_time(elapsed)
    s.poll

    assert_metrics_recorded({
      'CPU/User Time'   => { :call_count => 1, :total_call_time => 4.0 },
      'CPU/System Time' => { :call_count => 1, :total_call_time => 2.0 },
      # (4s user time)   / ((10s elapsed time) * 4 cpus) = 0.1
      'CPU/User/Utilization'   => { :call_count => 1, :total_call_time => 0.1 },
      # (2s system time) / ((10s elapsed time) * 4 cpus) = 0.05
      'CPU/System/Utilization' => { :call_count => 1, :total_call_time => 0.05 }
    })
  end

  def test_memory__default
    s = Mmtrix::Agent::Samplers::MemorySampler.new
    s.poll
    s.poll
    s.poll
    stats = @stats_engine.get_stats_no_scope("Memory/Physical")
    assert_equal(3, stats.call_count)
    assert stats.total_call_time > 0.5, "cpu greater than 0.5 ms: #{stats.total_call_time}"
  end

  def test_memory__linux
    return if RUBY_PLATFORM =~ /darwin/
    Mmtrix::Agent::Samplers::MemorySampler.any_instance.stubs(:platform).returns 'linux'
    s = Mmtrix::Agent::Samplers::MemorySampler.new
    s.poll
    s.poll
    s.poll
    stats = @stats_engine.get_stats_no_scope("Memory/Physical")
    assert_equal 3, stats.call_count
    assert stats.total_call_time > 0.5, "cpu greater than 0.5 ms: #{stats.total_call_time}"
  end

  def test_memory__solaris
    return if defined? JRuby
    Mmtrix::Agent::Samplers::MemorySampler.any_instance.stubs(:platform).returns 'solaris'
    Mmtrix::Agent::Samplers::MemorySampler::ShellPS.any_instance.stubs(:get_memory).returns 999
    s = Mmtrix::Agent::Samplers::MemorySampler.new
    s.poll
    stats = @stats_engine.get_stats_no_scope("Memory/Physical")
    assert_equal 1, stats.call_count
    assert_equal 999, stats.total_call_time
  end

  def test_memory__windows
    return if defined? JRuby
    Mmtrix::Agent::Samplers::MemorySampler.any_instance.stubs(:platform).returns 'win32'
    assert_raises Mmtrix::Agent::Sampler::Unsupported do
      Mmtrix::Agent::Samplers::MemorySampler.new
    end
  end

  def test_memory__is_supported
    Mmtrix::Agent::Samplers::MemorySampler.stubs(:platform).returns 'windows'
    assert !Mmtrix::Agent::Samplers::MemorySampler.supported_on_this_platform? || defined? JRuby
  end

end
