# encoding: utf-8
# This file is distributed under Mmtrix's license terms.


require File.expand_path(File.join(File.dirname(__FILE__),'..','test_helper'))
require 'mmtrix/agent/transaction/attributes'

class Mmtrix::Agent::NoticedErrorTest < Minitest::Test
  include Mmtrix::TestHelpers::Exceptions

  def setup
    @path = 'foo/bar/baz'

    freeze_time
    @time = Time.now

    @attributes = Mmtrix::Agent::Transaction::Attributes.new(Mmtrix::Agent.instance.attribute_filter)
    @attributes_from_notice_error = { :user => 'params' }
  end

  def test_to_collector_array
    e = TestError.new('test exception')

    error = create_error(e)
    error.request_uri = "http://com.google"
    error.attributes_from_notice_error = @attributes_from_notice_error

    expected = [
      (@time.to_f * 1000).round,
      @path,
      'test exception',
      'Mmtrix::TestHelpers::Exceptions::TestError',
      {
        :request_uri      => 'http://com.google',
        'userAttributes'  => { 'user' => 'params' },
        'agentAttributes' => {},
        'intrinsics'      => {}
      }
    ]
    assert_equal expected, error.to_collector_array
  end

  def test_to_collector_array_merges_custom_attributes_and_params
    e = TestError.new('test exception')
    @attributes.merge_custom_attributes(:custom => "attribute")

    error = create_error(e)
    error.attributes_from_notice_error = @attributes_from_notice_error

    actual = extract_attributes(error)
    expected = {
      'userAttributes' => {
        'user'   => 'params',
        'custom' => 'attribute'
      },
      'agentAttributes' => {},
      'intrinsics'      => {}
    }

    assert_equal expected, actual
  end

  def test_to_collector_array_includes_agent_attributes
    e = TestError.new('test exception')
    @attributes.add_agent_attribute(:agent, "attribute", Mmtrix::Agent::AttributeFilter::DST_ALL)
    error = create_error(e)

    actual = extract_attributes(error)
    assert_equal({:agent => "attribute"}, actual["agentAttributes"])
  end

  def test_to_collector_array_includes_intrinsic_attributes
    e = TestError.new('test exception')
    @attributes.add_intrinsic_attribute(:intrinsic, "attribute")
    error = create_error(e)

    actual = extract_attributes(error)
    assert_equal({:intrinsic => "attribute"}, actual["intrinsics"])
  end

  def test_to_collector_array_happy_without_attribute_collections
    error = Mmtrix::NoticedError.new(@path, "BOOM")

    expected = [
      (@time.to_f * 1000).round,
      @path,
      "BOOM",
      "Error",
      {
        'userAttributes'  => {},
        'agentAttributes' => {},
        'intrinsics'      => {}
      }
    ]
    assert_equal expected, error.to_collector_array
  end

  def test_to_collector_array_with_bad_values
    error = Mmtrix::NoticedError.new(@path, nil, Rational(10, 1))
    expected = [
      10_000.0,
      @path,
      "<no message>",
      "Error",
      {
        'userAttributes'  => {},
        'agentAttributes' => {},
        'intrinsics'      => {}
      }
    ]
    assert_equal expected, error.to_collector_array
  end

  def test_handles_non_string_exception_messages
    e = Exception.new({ :non => :string })
    error = Mmtrix::NoticedError.new(@path, e, @time)
    assert_equal(String, error.message.class)
  end

  def test_strips_message_from_exceptions_in_high_security_mode
    with_config(:high_security => true) do
      e = TestError.new('test exception')
      error = Mmtrix::NoticedError.new(@path, e, @time)

      assert_equal Mmtrix::NoticedError::STRIPPED_EXCEPTION_REPLACEMENT_MESSAGE, error.message
    end
  end

  def test_permits_messages_from_whitelisted_exceptions_in_high_security_mode
    with_config(:'strip_exception_messages.whitelist' => 'Mmtrix::TestHelpers::Exceptions::TestError') do
      e = TestError.new('whitelisted test exception')
      error = Mmtrix::NoticedError.new(@path, e, @time)

      assert_equal 'whitelisted test exception', error.message
    end
  end

  def test_whitelisted_returns_nil_with_an_empty_whitelist
    with_config(:'strip_exception_messages.whitelist' => '') do
      assert_falsy Mmtrix::NoticedError.passes_message_whitelist(TestError)
    end
  end

  def test_whitelisted_returns_nil_when_error_is_not_in_whitelist
    with_config(:'strip_exception_messages.whitelist' => 'YourErrorIsInAnotherCastle') do
      assert_falsy Mmtrix::NoticedError.passes_message_whitelist(TestError)
    end
  end

  def test_whitelisted_is_true_when_error_is_in_whitelist
    with_config(:'strip_exception_messages.whitelist' => 'OtherException,Mmtrix::TestHelpers::Exceptions::TestError') do
      assert_truthy Mmtrix::NoticedError.passes_message_whitelist(TestError)
    end
  end

  def test_whitelisted_ignores_nonexistent_exception_types_in_whitelist
    with_config(:'strip_exception_messages.whitelist' => 'NonExistent::Exception,Mmtrix::TestHelpers::Exceptions::TestError') do
      assert_truthy Mmtrix::NoticedError.passes_message_whitelist(TestError)
    end
  end

  def test_whitelisted_is_true_when_an_exceptions_ancestor_is_whitelisted
    with_config(:'strip_exception_messages.whitelist' => 'Mmtrix::TestHelpers::Exceptions::ParentException') do
      assert_truthy Mmtrix::NoticedError.passes_message_whitelist(ChildException)
    end
  end

  def test_handles_exception_with_nil_original_exception
    e = Exception.new('Buffy FOREVER')
    e.stubs(:original_exception).returns(nil)
    error = Mmtrix::NoticedError.new(@path, e, @time)
    assert_equal(error.message.to_s, 'Buffy FOREVER')
  end

  def test_intrinsics_always_get_sent
    with_config(:'error_collector.attributes.enabled' => false) do
      attributes = Mmtrix::Agent::Transaction::Attributes.new(Mmtrix::Agent.instance.attribute_filter)
      attributes.add_intrinsic_attribute(:intrinsic, "attribute")

      error = Mmtrix::NoticedError.new(@path, Exception.new("O_o"))
      error.attributes = attributes

      serialized_attributes = extract_attributes(error)
      assert_equal({ :intrinsic => "attribute" }, serialized_attributes["intrinsics"])
    end
  end

  def create_error(exception = StandardError.new)
    noticed_error = Mmtrix::NoticedError.new(@path, exception, @time)
    noticed_error.attributes = @attributes
    noticed_error
  end

  def extract_attributes(error)
    error.to_collector_array[4]
  end
end
